/*
 * Copyright (c) 1996-2015,2024 CNRS/LAAS
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
#include "acdemo.h"

#include <stdio.h>
#include <math.h>
#include "demoConst.h"

#include "demo_c_types.h"

#define SIGN(x) ((x)<0.?-1.0:1.0)

/* --- Task motion ------------------------------------------------------ */


/** Codel InitDemoSDI of task motion.
 *
 * Triggered by demo_start.
 * Yields to demo_ether.
 */
genom_event
InitDemoSDI(demo_ids *ids, const demo_Mobile *Mobile,
            const genom_context self)
{
  ids->state.position = 0.0;
  ids->state.speed = 0;
  Mobile->data(self)->position = ids->state.position;
  Mobile->data(self)->speed =   ids->state.speed;
  Mobile->write(self);		/* write the poster. */
  ids->speedRef = demo_SLOW;
  printf ("InitDemoSDI done.\n");
  return demo_ether;
}


/* --- Activity MoveDistance -------------------------------------------- */

/** Codel mdStartEngine of activity MoveDistance.
 *
 * Triggered by demo_start.
 * Yields to demo_exec, demo_ether.
 * Throws demo_TOO_FAR_AWAY.
 */
genom_event
mdStartEngine(double distRef, double position, double *posRef,
              const genom_context self)
{
  *posRef = position + distRef;
  printf("starting engine\n");
  return demo_exec;
}

/** Codel mdGotoPosition of activity MoveDistance.
 *
 * Triggered by demo_exec.
 * Yields to demo_pause_exec, demo_end.
 * Throws demo_TOO_FAR_AWAY.
 */
genom_event
mdGotoPosition(demo_speed speedRef, double posRef, demo_state *state,
               const demo_Mobile *Mobile, const genom_context self)
{
  double distRemain;
  double dDist;
  double v;

  distRemain = posRef - state->position;
  v = (speedRef == demo_SLOW) ?
    DEMO_DEFAULT_MAX_SPEED/4. : DEMO_DEFAULT_MAX_SPEED;
  dDist = v * demo_task_period * demo_millisecond* SIGN(distRemain);

  if (fabs(distRemain) < fabs(dDist)) {
       Mobile->data(self)->position = state->position = posRef;
       Mobile->data(self)->speed = state->speed = 0;
       Mobile->write(self);		/* write the poster. */
       return demo_end;
  }

  Mobile->data(self)->position = state->position += dDist;
  Mobile->data(self)->speed = state->speed = v;
  Mobile->write(self);		/* write the poster. */
#ifdef VERBOSE
  printf ("speed %g m/s  pos %g m\n", v, state->position);
#endif
  return demo_pause_exec;
}

/** Codel mdStopEngine of activity MoveDistance.
 *
 * Triggered by demo_end, demo_stop.
 * Yields to demo_ether.
 * Throws demo_TOO_FAR_AWAY.
 */
genom_event
mdStopEngine(const genom_context self)
{
  printf ("stop engine\n");
  return demo_ether;
}


/* --- Activity GotoPosition -------------------------------------------- */

/** Codel gpStartEngine of activity GotoPosition.
 *
 * Triggered by demo_start.
 * Yields to demo_exec, demo_ether.
 * Throws demo_TOO_FAR_AWAY.
 */
genom_event
gpStartEngine(const genom_context self)
{
  printf("starting engine\n");
  return demo_exec;
}

/** Codel gpGotoPosition of activity GotoPosition.
 *
 * Triggered by demo_exec.
 * Yields to demo_pause_exec, demo_end.
 * Throws demo_TOO_FAR_AWAY.
 */
genom_event
gpGotoPosition(double posRef, demo_ids *ids, const demo_Mobile *Mobile,
               const genom_context self)
{
  return mdGotoPosition(ids->speedRef, posRef, &ids->state, Mobile, self);
}

/** Codel gpStopEngine of activity GotoPosition.
 *
 * Triggered by demo_end, demo_stop.
 * Yields to demo_ether.
 * Throws demo_TOO_FAR_AWAY.
 */
genom_event
gpStopEngine(const genom_context self)
{
  return mdStopEngine(self);
}


/* --- Activity Track --------------------------------------------------- */

/** Codel gpStartEngine of activity Track.
 *
 * Triggered by demo_start.
 * Yields to demo_exec, demo_ether.
 * Throws demo_TOO_FAR_AWAY, demo_INVALID_REFERENCE.
 */
/* already defined in service GotoPosition */


/** Codel trackUpdateRef of activity Track.
 *
 * Triggered by demo_exec.
 * Yields to demo_goto.
 * Throws demo_TOO_FAR_AWAY, demo_INVALID_REFERENCE.
 */
genom_event
trackUpdateRef(const demo_Reference *Reference, demo_speed *speedRef,
               double *posRef, const genom_context self)
{
  const demo_reference *ref;
  genom_event e;

  /* update port contents */
  e = Reference->read(self);
  if (e != genom_ok) {
    gpStopEngine(self);
    return demo_INVALID_REFERENCE(self);
  }

  /* get port data */
  ref = Reference->data(self);
  if (ref == NULL) {
    gpStopEngine(self);
    return demo_INVALID_REFERENCE(self);
  }

  /* update IDS */
  *speedRef = ref->speed;
  *posRef = ref->position;

  return demo_goto;
}

/** Codel gpGotoPosition of activity Track.
 *
 * Triggered by demo_goto.
 * Yields to demo_pause_exec, demo_end.
 * Throws demo_TOO_FAR_AWAY, demo_INVALID_REFERENCE.
 */
/* already defined in service GotoPosition */


/** Codel gpStopEngine of activity Track.
 *
 * Triggered by demo_end, demo_stop.
 * Yields to demo_ether.
 * Throws demo_TOO_FAR_AWAY, demo_INVALID_REFERENCE.
 */
/* already defined in service GotoPosition */



/* --- Activity Monitor ------------------------------------------------- */

/** Codel monitor of activity Monitor.
 *
 * Triggered by demo_start.
 * Yields to demo_pause_start, demo_stop.
 * Throws demo_TOO_FAR_AWAY.
 */
genom_event
monitor(double monitor, const demo_ids *ids, const genom_context self)
{
  double dDist;

  dDist = ids->state.speed * demo_task_period * demo_millisecond;
  if (fabs (monitor - ids->state.position) < dDist) {
    printf ("dist %f mon %f pos %f\n", dDist, monitor, ids->state.position);
    return demo_stop;
  }
  return demo_pause_start;
}

/** Codel monitorStop of activity Monitor.
 *
 * Triggered by demo_stop.
 * Yields to demo_ether.
 * Throws demo_TOO_FAR_AWAY.
 */
genom_event
monitorStop(const demo_ids *ids, double *position,
            const genom_context self)
{
  *position = ids->state.position;
  return demo_ether;
}
